/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.bonsai.util.helper.Movements;
import de.unibi.airobots.btl.data.navigation.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.navigation.SpeedData;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;

import java.io.IOException;
import org.apache.log4j.Logger;

/**
 * Implementation of a simple follow person state.
 * 
 * @author nkoester
 */
public class IntroduceRobot extends AbstractState {
	/*
	 * What the robot will say.
	 */
	private static final String PRESS_LATER = "You may press them later.";
	private static final String EMERGENCY_BUTTONS = "I also have emergency buttons on both sides.";
	private static final String SENSORS_AND_ACTUATORS = "I have several sensors and actuators to explore and interact with "
			+ "the environment.";
	private static final String FIND_MY_WAY = "I can use prior maps to find my way, but also adapt to new "
			+ "environments easily.";
	private static final String CHALLENGE = "The major challenge for me is to learn about people and their "
			+ "homes by interaction.";

	/*
	 * Actuators used by this state.
	 */
	private SpeechActuator speechActuator;
	private NavigationActuator navActuator;

	/*
	 * Sensors used by this state.
	 */
	private Sensor<GlobalPlannerStateChange> gplanChangeSensor;
	private Sensor<PositionData> positionSensor;
	private Sensor<SpeedData> speedSensor;

	/**
	 * The logger.
	 */
	private Logger logger = super.getLogger();

	/**
	 * Constructor.
	 * 
	 * @param aStateMachine
	 *            The state machine to use.
	 */
	public IntroduceRobot(SkillStateMachine aStateMachine) {
		super(aStateMachine);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ExitStatus init() throws InitException {
		// Initialize actuators
		navActuator = getManagedActuator("NavigationActuator");
		speechActuator = getManagedActuator("SpeechActuator");

		// Initialize sensors
		positionSensor = getManagedSensor("PositionSensor");
		gplanChangeSensor = getManagedSensor("GlobalPlannerStateChangeSensor");
		speedSensor = getManagedSensor("SpeedSensor");

		say(CHALLENGE);
		say(FIND_MY_WAY);
		say(SENSORS_AND_ACTUATORS);

		try {
			sleep(1000);
		} catch (InterruptedException ex) {
			logger.warn(ex.getMessage());
		}

		PositionData positionData;
		try {
			positionData = positionSensor.readLast();

			NavigationGoalData rotateStart;

			rotateStart = CoordinateSystemConverter.polar2NavigationGoalData(
					positionData, (Math.PI / 2), 0.0, AngleUnit.RADIAN,
					LengthUnit.METER);

			rotateStart.setYawTolerance(Math.PI * 1 / 8, AngleUnit.RADIAN);
			rotateStart.setCoordinateTolerance(0.5, LengthUnit.METER);

			driveToPosition(rotateStart, GoalType.LOCAL);
			say(EMERGENCY_BUTTONS);
			say(PRESS_LATER);

			positionData = positionSensor.readLast();

			NavigationGoalData rotateEnd;
			rotateEnd = CoordinateSystemConverter.polar2NavigationGoalData(
					positionData, (-Math.PI / 2), 0.0, AngleUnit.RADIAN,
					LengthUnit.METER);

			rotateEnd.setYawTolerance(Math.PI / 1 / 8, AngleUnit.RADIAN);
			rotateEnd.setCoordinateTolerance(0.5, LengthUnit.METER);

			driveToPosition(rotateEnd, GoalType.LOCAL);
		} catch (InterruptedException ex) {
			throw new InitException(ex.getMessage());
		} catch (IOException ex) {
			throw new InitException(ex.getMessage());
		}

		return ExitStatus.SUCCESS;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ExitStatus execute() throws ExecuteException {
		return ExitStatus.SUCCESS;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public ExitStatus end() throws EndException {
		return ExitStatus.SUCCESS;
	}

	public boolean driveToPosition(NavigationGoalData target, GoalType type)
			throws IOException {
		return Movements.driveToPosition(navActuator, positionSensor,
				speedSensor, gplanChangeSensor, target, type);
	}

	/**
	 * Use speech actuator to say something and catch IO exception.
	 * 
	 * @param text
	 *            Text to be said.
	 */
	private void say(String text) {
		try {
			speechActuator.say(text);
		} catch (IOException ex) {
			// Not so bad. The robot just says nothing.
			logger.warn(ex.getMessage());
		}
	}
}
