/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.states;

import de.unibi.airobots.bonsai.actuators.BtlMemoryActuator;
import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.SpeechActuator;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.FaceIdentificationSensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.skills.SkillStateMachine;
import de.unibi.airobots.bonsai.skills.abstractstates.AbstractState;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.EndException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.ExecuteException;
import de.unibi.airobots.bonsai.skills.abstractstates.exceptions.InitException;
import de.unibi.airobots.bonsai.skills.helper.PersonManager;
import de.unibi.airobots.btl.data.navigation.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.navigation.SpeedData;
import de.unibi.airobots.btl.data.person.PersonData;
import de.unibi.airobots.btl.data.person.PersonDataList;
import java.io.IOException;
import java.util.ArrayList;
import org.apache.log4j.Logger;

/**
 * Use this state to recognize all persons in front of the robot (maximum 2). 
 * 
 * ToDo: implement correctly and document!!!
 * 
 * @author lkettenb
 */
public class RecognizePersons extends AbstractState {
    /*
     * What the robot will say.
     */
    private static final String STATE_TWOPERSONS = 
            "There are two persons in front of me.!";
    private static final String STATE_RECOGNIZE = 
            "Let me check if know one of them!";
    private static final String STATE_FINISHED = 
            "My owner is back! That makes me happy!";
    
    /*
     * Actuators used by this state.
     */
    private SpeechActuator speechActuator;
    private BtlMemoryActuator memoryActuator;
    private NavigationActuator navigationActuator;
    
    /*
     * Sensors used by this state.
     */
    private FaceIdentificationSensor faceSensor;
    private Sensor<PersonList> personSensor;
    private Sensor<PersonDataList> personDataListSensor;
    private Sensor<PositionData> positionSensor;
    private Sensor<SpeedData> speedSensor;
    private Sensor<GlobalPlannerStateChange> gplanChangeSensor;
    
    /**
     * List with person data. Will be read from and saved to the scene memory.
     */
    private PersonDataList personDataList;    
    
    /**
     * The logger.
     */
    private Logger logger = getLogger();
    
    /**
     * Time to sleep while waiting.
     */
    private static final int TIME_TO_SLEEP = 350;
    
    /**
     * List of persons in front of the robot.
     */
    private ArrayList<PersonData> cleanedList = new ArrayList<PersonData>();
    
    
    /**
     * Constructor.
     * @param aStateMachine The state machine to use.
     */
    public RecognizePersons(SkillStateMachine aStateMachine) {
        super(aStateMachine);
    }

    
    @Override
    public ExitStatus init() throws InitException {
        // Initialize actuators
        speechActuator = getManagedActuator("SpeechActuator");
        memoryActuator = getManagedActuator("BtlMemoryActuator");
        navigationActuator = getManagedActuator("NavigationActuator");
        // Initialize sensors
        faceSensor = (FaceIdentificationSensor)
                getManagedSensor("FaceIdentificationSensor");
        personSensor = getManagedSensor("PersonSensor");
        personDataListSensor = getManagedSensor("BtlMemorySensor");
        positionSensor = getManagedSensor("PositionSensor");
        speedSensor = getManagedSensor("SpeedSensor");
        gplanChangeSensor = getManagedSensor("GlobalPlannerStateChangeSensor");
        
        logger.debug("Unable to recognize person. "
                + "No person data was written to SceneMemory.");
        return ExitStatus.ERROR;
    }

    
    @Override
    public ExitStatus execute() throws ExecuteException {
        // waiting for two persons in front
        logger.info("Wait until two persons in front of the robot!"); 
        
        PersonList list;
        try {
            list = personSensor.readLast();
        } catch (IOException ex) {
            throw new ExecuteException(ex.getMessage());
        } catch (InterruptedException ex) {
            throw new ExecuteException(ex.getMessage());
        }
        
        cleanedList.clear();
        for (PersonData p : list) {
            if (PersonManager.isStandingInFront(p)) {
                cleanedList.add(p);
            }
        }
        if (cleanedList.size() < 2) {
            try {
                sleep(TIME_TO_SLEEP);
            } catch (InterruptedException ex) {
                logger.warn(ex.getMessage());
            }
            setProcessingStatus(getWaitProcessingMessage());
            return ExitStatus.SUCCESS;
        }
        
        logger.info("Two persons in front of the robot!");
       
        // trying to recognize the owner
        say(STATE_TWOPERSONS);
        say(STATE_RECOGNIZE);
        
//        // cropping of the left of the two faces
//        faceSensor.twoFaces();
//        
//        // try to recognize left person
//        int recognizedID = PersonManager.recognize(faceSensor);
//        
//        // stop cropping of the left of the two faces
//        faceSensor.twoFaces();
        
        
        // drive to person
        boolean reachedPerson = PersonManager.driveToPerson(list.getPersonById(0), 
                navigationActuator, positionSensor, 
                speedSensor, gplanChangeSensor);
        
        say(STATE_FINISHED);
        
        return ExitStatus.SUCCESS;
    }

    
    @Override
    public ExitStatus end() throws EndException {
        return ExitStatus.SUCCESS;
    }    
    
    /**
     * Use speech actuator to say something and catch IO exception.
     * 
     * @param text Text to be said.
     */
    private void say(String text) {
        try {
            speechActuator.say(text);
        } catch (IOException ex) {
            // Not so bad. The robot just says nothing.
            logger.warn(ex.getMessage());
        }
    }
}
