/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.skills.view;

import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JPanel;

/**
 * A simple panel to control a SCXML state machine. Show currently possible 
 * events and lets the user start, restart and stop the state machine.
 * 
 * @author lkettenb
 */
public class StateMachineControlPanel extends JPanel implements ActionListener {
    
    /**
     * Action command fired when 'send' button is pressed.
     * 
     * @see StateMachineControlPanel#getCurrentSelectedEvent() 
     */
    public static final String AC_FIRE_EVENT = "AC_FIRE_EVENT";
    
    /**
     * Action command fired when 'start' button is pressed.
     */
    public static final String AC_START_EVENT = "AC_START_EVENT";
    
    /*
     * Buttons used by this panel. 
     */
    private JButton bFireButton = new JButton("Send");
    private JButton bStart = new JButton("Start");
    
    /*
     * ComboBoxe with possible events.
     */
    private JComboBox cbEventList;
    
    /**
     * This action listener will be notified to control the state machine.
     */
    private ActionListener actionListener;
    
    
    /**
     * Convenience constructor.
     *
     * @param actionListener ActionListener used by this panel.
     * 
     * @see StateMachineControlPanel#actionListener
     * @see StateMachineControlPanel#AC_FIRE_EVENT
     * @see StateMachineControlPanel#AC_START_EVENT
     */
    public StateMachineControlPanel(ActionListener actionListener) {
        super();
        this.actionListener = actionListener;
        init();
    }
    
    
    /**
     * Initializes everything.
     */
    private void init() {
        setLayout(new FlowLayout(FlowLayout.CENTER));        
        
        String[] events = { "Start engine" };
        cbEventList = new JComboBox(events);
        cbEventList.setEnabled(false);
        
        add(cbEventList);
        
        bFireButton.addActionListener(actionListener);
        bFireButton.setActionCommand(AC_FIRE_EVENT);
        bFireButton.setEnabled(false);
        add(bFireButton);
        
        bStart.addActionListener(actionListener);
        bStart.addActionListener(this);
        bStart.setActionCommand(AC_START_EVENT);
        add(bStart);
        
        setBorder(BorderFactory.createTitledBorder(" Controller "));
    }
    
    
    /**
     * The currently selected (validated) event.
     * 
     * @return Currently selected (validated) event.
     */
    public String getCurrentSelectedEvent() {
        return cbEventList.getSelectedItem().toString();
    }
    
    
    @Override
    public void setEnabled(boolean enable) {
        bStart.setEnabled(enable);
    }
    
    
    /**
     * Update list of currently possible events.
     * 
     * @param eventList List of currently possible events.
     */
    public void updateEventList(String[] eventList) {
        cbEventList.removeAllItems();
        for (String event : eventList) {
            cbEventList.addItem(event);
        }
    }

    
    @Override
    public void actionPerformed(ActionEvent e) {
        if (e.getActionCommand().equals(AC_START_EVENT)) {
            bFireButton.setEnabled(true);
            cbEventList.setEnabled(true);
            bStart.setEnabled(false);
        }
    }
}
