/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.strategies.follow;

import de.unibi.airobots.bonsai.actuators.NavigationActuator;
import de.unibi.airobots.bonsai.actuators.TaskPatternActuator;
import de.unibi.airobots.bonsai.core.BonsaiManager;
import de.unibi.airobots.bonsai.core.Sensor;
import de.unibi.airobots.bonsai.sensors.data.PersonList;
import de.unibi.airobots.bonsai.util.CoordinateSystemConverter;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter;
import de.unibi.airobots.bonsai.util.helper.GoalReachedWaiter.ReachStatus;
import de.unibi.airobots.bonsai.xcf.actuators.XcfTaskPatternActuator;
import de.unibi.airobots.btl.data.navigation.GlobalPlannerStateChange;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData.GoalType;
import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.airobots.btl.data.navigation.SpeedData;
import de.unibi.airobots.btl.data.person.PersonData;
import de.unibi.airobots.btl.data.taskpattern.PersonFoundNotification;
import de.unibi.airobots.btl.data.taskpattern.PleaseWaitNotification;
import de.unibi.airobots.btl.data.taskpattern.RobotLostHumanNotification;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;

import java.io.IOException;
import java.util.LinkedList;
import net.sf.xcf.transport.XOPData;
import nu.xom.Builder;
import nu.xom.ParsingException;
import nu.xom.ValidityException;
import org.apache.log4j.Logger;

public class TaskPatternBasedSimpleFollowStrategy implements FollowStrategy {

	private Logger logger = Logger
			.getLogger(TaskPatternBasedSimpleFollowStrategy.class);

	private static final double MAX_PERCEPTION_AGE = 2000;

	private static final double DISTANCE_THRESHOLD_STOP = 1000;

	private static final double DISTANCE_THRESHOLD_MOVE = 5000;

	private static final double DISTANCE_THRESHOLD_TURN = 1000;

	private static final double DISTANCE_THRESHOLD_START_FOLLOWING = 3000;

	private static final double RESUME_MAX_DIST = 1900;

	private static final double COORDINATE_TOLERANCE = 0.5;
	private static final double YAW_TOLERANCE = Math.PI * 0.3;

	private static final int QUEUE_LENGTH = 10;

	private int personFollowId = -1;

	private int noPersonCounter = 0;

	private LinkedList<NavigationGoalData> lastKnownPositions = new LinkedList<NavigationGoalData>();

	private volatile boolean firstTimeLost = true;
	private volatile boolean stepForwardAllowed = true;
	private int forewardSteps = 0;
	private static final double STEP_FORWARD = 1.0;
	private static final int MAX_FOREWARD_STEPS = 3;

	Sensor<PersonList> personSensor;
	Sensor<PositionData> posSensor;
	Sensor<SpeedData> speedSensor;
	NavigationActuator navActuator;
	TaskPatternActuator taskActuator;
	Sensor<GlobalPlannerStateChange> gplanChangeSensor;

	private Builder builder = new Builder();

	public TaskPatternBasedSimpleFollowStrategy() {
		/*
		 * try { BonsaiManager.getInstance().configure(
		 * GoGetItRunner.class.getResource("../xml/global.xml") .toURI()); }
		 * catch (IllegalStateException e) { // TODO Auto-generated catch block
		 * e.printStackTrace(); } catch (ConfigurationException e) { // TODO
		 * Auto-generated catch block e.printStackTrace(); } catch
		 * (URISyntaxException e) { // TODO Auto-generated catch block
		 * e.printStackTrace(); }
		 */

		personSensor = BonsaiManager.getInstance().createSensor("PersonSensor",
				PersonList.class);
		posSensor = BonsaiManager.getInstance().createSensor("PositionSensor",
				PositionData.class);
		taskActuator = BonsaiManager.getInstance().createActuator(
				"TaskPatternActuator", XcfTaskPatternActuator.class);
		navActuator = BonsaiManager.getInstance().createActuator(
				"NavigationActuator", NavigationActuator.class);
		speedSensor = BonsaiManager.getInstance().createSensor("SpeedSensor",
				SpeedData.class);
		gplanChangeSensor = BonsaiManager.getInstance().createSensor(
				"GlobalPlannerStateChangeSensor",
				GlobalPlannerStateChange.class);
	}

	@Override
	public void init() {
		personFollowId = -1;
		lastKnownPositions.clear();

	};

	@Override
	public void execute() throws IOException {

		stepForwardAllowed = true;
		forewardSteps = 0;

		logger.debug("check for person");

		PersonList personList;
		try {
			personList = personSensor.readLast();

		} catch (InterruptedException e1) {
			logger.fatal(e1.getMessage());
			return;
		}

		PersonData personFollow = null;

		// try to find followed person.
		if (personFollowId == -1) {

			personFollow = searchNewPerson(personList);

		} else {

			personFollow = searchKnownPerson(personList);

		}
		// If a followed person was defined.
		logger.debug("person Follow " + personFollow);
		if (personFollow != null) {
			follow(personFollow);
		}

	}

	@Override
	public void execute(int personId) throws IOException {

		logger.debug("check for person with id " + personId);

		PersonList personList;
		try {
			personList = personSensor.readLast();
			logger.debug("personList " + personList);
		} catch (InterruptedException e1) {
			logger.fatal(e1.getMessage());
			return;
		}

		personFollowId = personId;

		PersonData personFollow = null;

		// try to find followed person.
		if (personFollowId == -1) {
			personFollow = searchNewPerson(personList);
		} else {
			personFollow = searchKnownPerson(personList);
		}

		logger.debug("person Follow " + personFollow);

		// If a followed person was defined.
		if (personFollow != null) {
			follow(personFollow);
		}
	}

	private PersonData searchNewPerson(PersonList personList)
			throws IOException {

		PersonData personFollow = null;

		// choose nearest person
		double minDist = Double.MAX_VALUE;

		for (PersonData p : personList) {

			boolean isNear = p.getDistance(LengthUnit.MILLIMETER) <= DISTANCE_THRESHOLD_START_FOLLOWING;

			boolean isSmallId = p.getId() >= 0;
			// adding a penality for greater person angles so persons in front
			// of the robot are prefered.
			double pdist = p.getDistance(LengthUnit.MILLIMETER)
					* (Math.abs(p.getAngle(AngleUnit.RADIAN)) + 1.0);

			if (pdist < minDist && isNear && isSmallId) {
				minDist = p.getDistance(LengthUnit.MILLIMETER);
				personFollow = p;
			}
		}

		if (personFollow != null) {

			personFollowId = personFollow.getId();
			logger.debug("Following person [" + personFollowId + "]");

		} else {

			if (stepForwardAllowed) {
				logger.debug("stepForward in searchNewPerson");
				stepForward();
				if (forewardSteps >= MAX_FOREWARD_STEPS) {
					stepForwardAllowed = false;
				} else {
					forewardSteps++;
				}
			}
			if (firstTimeLost && !lastKnownPositions.isEmpty()) {
				// logger.debug("step forward?!");
				try {
					taskActuator.initiate(new XOPData(builder.build("<"
							+ RobotLostHumanNotification.getBaseTag() + "/>",
							null)));
				} catch (ValidityException e) {
					logger.error(e);
				} catch (ParsingException e) {
					logger.error(e);
				}

				firstTimeLost = false;
			}
			logger.debug("No person detected");
			noPersonCounter++;
		}

		return personFollow;

	}

	private void stepForward() {
		PositionData positionData;
		try {
			positionData = posSensor.readLast();
			NavigationGoalData g;

			g = CoordinateSystemConverter.polar2NavigationGoalData(
					positionData, 0, STEP_FORWARD, AngleUnit.RADIAN,
					LengthUnit.MILLIMETER);

			g.setYawTolerance(YAW_TOLERANCE, AngleUnit.RADIAN);
			g.setCoordinateTolerance(COORDINATE_TOLERANCE, LengthUnit.METER);
			g.setType(GoalType.LOCAL);

			GoalReachedWaiter waiter = new GoalReachedWaiter(posSensor,
					speedSensor, gplanChangeSensor, navActuator, 800,
					GoalReachedWaiter.DEFAULT_STEPS);
			navActuator.setGoal(g);
			ReachStatus goalReached = waiter.waitForReachingGoal(g, 1000);
			logger.debug("goalReached: " + goalReached);
		} catch (InterruptedException e) {
			logger.warn("Interrupted: " + e);
		} catch (IOException e) {
			logger.warn("IOException: " + e);
		}
	}

	private PersonData searchKnownPerson(PersonList personList)
			throws IOException {

		if (personList == null)
			try {
				personList = personSensor.readLast();
				logger.debug(personList);
			} catch (InterruptedException e1) {
				logger.warn(e1);
			}

		PersonData personFollow = personList.getPersonById(personFollowId);
		logger.info("personData " + personFollow);
		String list = "";
		for (PersonData d : personList) {
			list += " " + d.getId();
		}

		logger.info("following person " + personFollow + ". Found IDs in scan:"
				+ list);

		// check if percept too old
		// if (personFollow != null
		// && System.currentTimeMillis()
		// - personFollow.getTimestamp().getUpdated().getTime() >
		// MAX_PERCEPTION_AGE) {
		// personFollow = null;
		// }

		// search for followed person, if not in person list.
		if (personFollow == null) {

			try {
				PositionData currentPosition = posSensor.readLast();

				// choose position
				// TODO check m / mm
				double minDist2 = Double.MAX_VALUE;
				int nearestId = -1;
				for (NavigationGoalData d : lastKnownPositions) {
					double dist = Math.hypot(
							currentPosition.getX(LengthUnit.METER)
									- d.getX(LengthUnit.METER),
							currentPosition.getY(LengthUnit.METER)
									- d.getY(LengthUnit.METER));
					if (dist < minDist2) {
						minDist2 = dist;
						nearestId = lastKnownPositions.indexOf(d);
					}
				}

				/*
				 * int desiredId = Math .round((lastKnownPositions.size() - 1 -
				 * nearestId) / 2) + nearestId;
				 * 
				 * // quick fix desiredId--;
				 * 
				 * logger.debug("chose position id " + desiredId +
				 * " for searchingposition.");
				 */

				try {
					taskActuator
							.initiate(new XOPData(builder.build("<"
									+ PleaseWaitNotification.getBaseTag()
									+ "/>", null)));
				} catch (ValidityException e) {
					logger.error(e);
				} catch (ParsingException e) {
					logger.error(e);
				}

				NavigationGoalData lastKnownPosition = lastKnownPositions
						.get(nearestId);

				if (lastKnownPosition != null) {
					logger.debug("go to lastKnownPosition");
					lastKnownPosition.setCoordinateTolerance(
							COORDINATE_TOLERANCE, LengthUnit.METER);
					lastKnownPosition.setYawTolerance(YAW_TOLERANCE,
							AngleUnit.RADIAN);
					GoalReachedWaiter waiter = new GoalReachedWaiter(posSensor,
							speedSensor, gplanChangeSensor, navActuator, 800,
							GoalReachedWaiter.DEFAULT_STEPS);

					navActuator.setGoal(lastKnownPosition);
					ReachStatus goalReached = waiter.waitForReachingGoal(
							lastKnownPosition, 10000);
					if (goalReached == ReachStatus.REACHED) {
						logger.debug("last known position reached");
					} else {
						logger.warn("---> last known position not reached <--- ("
								+ goalReached + ")");
					}
				}
			} catch (IndexOutOfBoundsException e) {
				logger.warn("Desired position id not available.");
			} catch (InterruptedException e) {
				logger.warn(e.getMessage());
			}

			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				logger.warn("Interrupted while waiting");
			}

			try {
				personList = personSensor.readLast();
				personFollow = personList.getPersonById(personFollowId);
				if (personList != null && personFollowId != -1
						&& personFollow != null) {
					logger.debug("get person " + personList);
					return personFollow;
				}

				if (personList.size() != 0) {

					// choose nearest person
					double minDist = Double.MAX_VALUE;
					for (PersonData p : personList) {

						// boolean isYoung = System.currentTimeMillis()
						// - p.getTimestamp().getUpdated().getTime() <
						// MAX_PERCEPTION_AGE;
						boolean isSmallId = p.getId() < 1000 && p.getId() >= 0;

						if (p.getDistance(LengthUnit.MILLIMETER) < minDist
								&& isSmallId) {
							minDist = p.getDistance(LengthUnit.MILLIMETER);
							personFollow = p;
						}
					}

					if (personFollow != null
					/* && personFollow.getDistance() < RESUME_MAX_DIST */
					&& noPersonCounter > 5) {
						try {
							taskActuator.initiate(new XOPData(builder.build("<"
									+ PersonFoundNotification.getBaseTag()
									+ "/>", null)));
						} catch (ValidityException e) {
							logger.error(e);
						} catch (ParsingException e) {
							logger.error(e);
						}
					} else {
						personFollow = null;
						// speechActuator.say("I can't see you anymore. "
						// + "I will follow someone else.");
						noPersonCounter++;
					}
				}
			} catch (IOException e) {
				logger.fatal(e.getMessage());
			} catch (InterruptedException e) {
				logger.fatal(e.getMessage());
			}

			personFollowId = -1;

		}
		return personFollow;
	}

	private void follow(PersonData personFollow) throws IOException {

		// logger.info(personFo);
		noPersonCounter = 0;
		logger.debug("Following Person: " + personFollow);

		double personAngle = personFollow.getAngle(AngleUnit.RADIAN);
		double personDist = personFollow.getDistance(LengthUnit.MILLIMETER);

		PositionData currentPosition;
		try {

			currentPosition = posSensor.readLast();
			logger.debug("currentPosition: "
					+ currentPosition.getX(LengthUnit.METER) + " / "
					+ currentPosition.getY(LengthUnit.METER));

			// System.out.println(personDist);

			if (personDist < DISTANCE_THRESHOLD_MOVE
					&& personDist > DISTANCE_THRESHOLD_TURN) {
				logger.debug("personDist< DTM && pD>DTT");
				NavigationGoalData personGoal = CoordinateSystemConverter
						.polar2NavigationGoalData(currentPosition, personAngle,
								(personDist - DISTANCE_THRESHOLD_STOP),
								AngleUnit.RADIAN, LengthUnit.MILLIMETER);

				if (lastKnownPositions.size() >= QUEUE_LENGTH) {
					lastKnownPositions.poll();
				}
				lastKnownPositions.add(personGoal);

				/*
				 * logger.info("Approaching person [" + personFollow.getId() +
				 * "] (Dist: " + personDist + ")");
				 * 
				 * logger .debug("Person NavigationGoal is x=" +
				 * personGoal.getX() + " y=" + personGoal.getY() + " yaw=" +
				 * personGoal.getYaw());
				 */
				logger.debug("Moving: "
						+ (personDist - DISTANCE_THRESHOLD_STOP) + "m"
						+ "with angle: " + personAngle + " (Radiant)");
				// String currentX = Double.toString(currentPosition.getX())
				// .substring(0, 4);
				// String currentY = Double.toString(currentPosition.getY())
				// .substring(0, 4);
				// String newX = Double.toString(personGoal.getX())
				// .substring(0, 4);
				// String newY = Double.toString(personGoal.getY())
				// .substring(0, 4);
				// logger.debug("Current position: x:" + currentX + " y:"
				// + currentY);
				// logger.debug("New position: x:" + newX + " y:" + newY);

				// personGoal.setYaw(personAngle);

				personGoal.setCoordinateTolerance(COORDINATE_TOLERANCE,
						LengthUnit.METER);
				personGoal.setYawTolerance(YAW_TOLERANCE, AngleUnit.RADIAN);
				navActuator.setGoal(personGoal);

			} else if (personDist < DISTANCE_THRESHOLD_TURN) {

				NavigationGoalData personGoal = CoordinateSystemConverter
						.polar2NavigationGoalData(currentPosition, personAngle,
								0.0, AngleUnit.RADIAN, LengthUnit.MILLIMETER);

				/*
				 * logger .info("Turning towards person [" +
				 * personFollow.getId() + "]");
				 */
				logger.debug("Turning: " + personAngle + " (Radiant)");
				personGoal.setCoordinateTolerance(COORDINATE_TOLERANCE,
						LengthUnit.METER);
				personGoal.setYawTolerance(YAW_TOLERANCE, AngleUnit.RADIAN);

				navActuator.setGoal(personGoal);
			} else {
				if (stepForwardAllowed) {
					logger.info("stepForward in following");
					stepForward();
				}
			}
		} catch (InterruptedException e) {
			logger.fatal("Don't know where I am. " + e.getMessage());
		}
	}

}
