/*
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 *
 * Copyright(c) Frederic Siepmann
 * http://opensource.cit-ec.de/projects/bonsai
 *
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 *
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 *
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 *
 * Contributors: Leon Ziegler, Johannes Wienke, Norman Köster, Torben Töniges,
 *               Annika Rothert, Alexander Baumann, Marcel Dulisch, Maikel Linke,
 *               Marian Pohling, Lukas Twardon, Lukas Kettenbach
 */
package de.unibi.airobots.bonsai.xcf.actuators;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.AsyncCallback;
import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.memory.DocumentNotFoundException;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.memory.SimpleMessage;
import net.sf.xcf.transport.Attachment;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.XPath;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Elements;

import org.apache.log4j.Logger;

import de.unibi.airobots.bonsai.actuators.SceneActuator;
import de.unibi.airobots.bonsai.util.DataConverter;
import de.unibi.airobots.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.Type.NoSourceDocumentException;
import de.unibi.airobots.btl.data.map.DynamicGridMap;
import de.unibi.airobots.btl.data.object.ObjectPositionData;

/**
 * "Xcf Active Memory" based actuator to control a scene memory containing meta
 * information and semantic annotations of the scene.
 * 
 * @author lziegler
 */
public class XcfSceneActuator implements SceneActuator,
        ActiveMemoryReconnectable, RemoteServerReconnectable {

    private Logger logger = Logger.getLogger(getClass());

    private String memoryName = "";
    private Object memoryLock = new Object();
    private ActiveMemory memory = null;
    private RemoteServer memoryServer = null;
    private int removeCount = 0;
    private boolean doCompacting = false;

    /**
     * Defines how many deletions are required before compact() will be called.
     */
    private static final int COMPACT_REMOVE_THRESHOLD = 100;

    /**
     * Xml namespace for active memory documents.
     */
    public static final String NAMESPACE_MEMORY =
            "http://purl.org/ActiveMemory";
    /**
     * Xml namespace for active memory attachment descriptions.
     */
    public static final String NAMESPACE_DESCRIPTION =
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
    /**
     * Xml namespace for dbxml.
     */
    public static final String NAMESPACE_DBXML =
            "http://www.sleepycat.com/2002/dbxml";

    /**
     * Creates a new instance of {@link XcfSceneActuator}.
     * 
     * @param memoryName
     *            Name of the memory to connect to.
     */
    public XcfSceneActuator(String memoryName) {
        this.memoryName = memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {
        synchronized (memoryLock) {
            this.memory = activeMemory;
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @throws IOException
     */
    @Override
    public void addObject(ObjectPositionData object) throws IOException {
        if (memory == null) {
            logger.error("Scene memory is not reachable");
        }
        try {
            memory.insert(new XOPData(object
                    .toDocument(ObjectPositionData.class)));
        } catch (MemoryException e) {
            logger.error("Scene memory is not reachable");
            logger.debug(e);
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     * 
     * @throws IOException
     */
    @Override
    public void updateObject(ObjectPositionData object)
            throws NoSourceDocumentException, IOException {
        if (memory == null) {
            logger.error("Scene memory is not reachable");
        }
        try {
            // copy source document
            Element source = new Element(object.getSourceDocument());

            // fill source document with new data
            object.fillInto(source);
            memory.replace(new XOPData(new Document(source)));

        } catch (DocumentNotFoundException e) {
            logger.error("ObjectPositionData can not be updated. "
                    + "Corresponding document not found in memory.");
            logger.debug(e);
            throw new IOException(e);
        } catch (MemoryException e) {
            logger.error("Scene memory is not reachable");
            logger.debug(e);
            throw new IOException(e);
        }
    }

    @Override
    public void addGridMap(DynamicGridMap map) throws IOException {
        if (memory == null) {
            logger.error("Scene memory is NULL.");
        }
        try {
            XOPData data =
                    new XOPData(map.toDocument(DynamicGridMap.class));

            // create attachment
            Attachment attachment = new Attachment();
            attachment.setValue(DataConverter.toByteArray(map.getLinearMap()));
            data.addAttachment(map.getUri(), attachment);
            memory.insert(data);
        } catch (MemoryException e) {
            logger.error("Scene memory Exception");
            logger.debug(e);
            throw new IOException(e);
        }
    }

    @Override
    public void removeGridMap(DynamicGridMap map)
            throws NoSourceDocumentException, IOException {

        if (memory == null) {
            logger.error("Scene memory is not reachable");
        }
        try {
            // copy source document
            Element source = new Element(map.getSourceDocument());

            // delete old data
            String path =
                    "//" + DynamicGridMap.getBaseTag() + "[@uri='"
                            + map.getUri() + "']";
            memory.remove(new XPath(path));
            removeCount++;

            // delete old attachment description from document
            Elements ms =
                    source.getChildElements("description",
                            NAMESPACE_DESCRIPTION);
            for (int i = 0; i < ms.size(); i++) {
                source.removeChild(ms.get(i));
            }

            map.setSourceDocument(source);

            // check if compact should be called
            if (removeCount >= COMPACT_REMOVE_THRESHOLD) {
                removeCount = 0;
                if (doCompacting) {
                    compactDatabase();
                }
            }

        } catch (DocumentNotFoundException e) {
            logger.error("DynamicGridMapData can not be removed. "
                    + "Corresponding document not found in memory.");
            logger.debug(e);
            throw new IOException(e);
        } catch (MemoryException e) {
            logger.error("Scene memory is not reachable");
            logger.debug(e);
            throw new IOException(e);
        }

    }

    /**
     * Enable or disable the database compacting.
     * 
     * @param enable
     *            If true, the database will be compacted frequently.
     */
    public void enableCompactingRoutine(boolean enable) {
        doCompacting = enable;
    }

    /**
     * This method sends a signal to the memory, that compacts the database on
     * the disk.
     */
    public void compactDatabase() {

        logger.info("compacting database");
        if (memoryServer != null) {
            try {
                // create the required document body
                SimpleMessage msg = new SimpleMessage();
                msg.setContent("1");

                // call the method on the memory server
                memoryServer.callMethod("compact",
                        new XOPData(msg.getAsDocument()), new AsyncCallback() {

                            @Override
                            public void error(Throwable arg0) {
                                logger.error("Exception while calling 'compact()' on memory server.");
                                logger.debug(arg0);
                                logger.debug(arg0.getCause());
                            }

                            @Override
                            public void complete(XOPData arg0) {
                                // done
                            }
                        });
            } catch (MethodNotFoundException e) {
                logger.error("Method 'compact()' not found on memory server. "
                        + "Wrong version of ActiveMemory installed?");
            } catch (UserException e) {
                logger.error("UserException while calling 'compact()' on memory server.");
                logger.debug(e);
            }
        } else {
            logger.warn("Memory remote server is not set. Cannot call compact method!");
        }
    }

    @Override
    public void updateGridMap(DynamicGridMap map)
            throws NoSourceDocumentException, IOException {

        if (memory == null) {
            logger.error("Scene memory is not reachable");
        }
        if (map == null) {
            logger.error("given map was null.");
            return;
        }
        if (map != null && (map.getGridMap() == null || map.getGridMap().length == 0)) {
            logger.error("given container has no containing grid map");
            return;
        }
        try {
            // copy source document
            Element source = new Element(map.getSourceDocument());

            // fill source document with new data
            map.fillInto(source);
            XOPData data = new XOPData(new Document(source));

            // create attachment
            Attachment attachment = new Attachment();
            attachment.setValue(DataConverter.toByteArray(map.getLinearMap()));
            data.addAttachment(map.getUri(), attachment);

            // publish under new id
            memory.replace(data);

        } catch (DocumentNotFoundException e) {
            logger.error("DynamicGridMapData can not be updated. "
                    + "Corresponding document not found in memory.");
            logger.debug(e);
            throw new IOException(e);
        } catch (MemoryException e) {
            logger.error("Scene memory is not reachable");
            logger.debug(e);
            throw new IOException(e);
        }
    }

    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        HashMap<String, ErrorHandlingPolicy> map =
                new HashMap<String, RemoteServerReconnectable.ErrorHandlingPolicy>();
        map.put(memoryName, ErrorHandlingPolicy.SELF_HEALING);
        return map;
    }

    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        memoryServer = remoteServer;
    }
}
